unit main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, ExtCtrls, Buttons, StdCtrls, mwFastTime,DXPath;

Const
      MapW   = 80;
      MapH   = 60;
      dL     = 8;


type
  TMainForm = class(TForm)
    Paint: TPaintBox;
    ObstacleButton: TSpeedButton;
    StartButton: TSpeedButton;
    FinishButton: TSpeedButton;
    PathButton: TButton;
    RefreshButton: TButton;
    MyStart1Check: TCheckBox;
    Label1: TLabel;
    Label2: TLabel;
    mwFastTime1: TmwFastTime;
    Button1: TButton;
    Button2: TButton;
    procedure FormCreate(Sender: TObject);
    procedure PaintPaint(Sender: TObject);
    procedure PaintMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PathButtonClick(Sender: TObject);
    procedure RefreshButtonClick(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
  private
    DimW,DimH : Integer;
    Drawing:Byte;
    procedure LoadArrFromStream(s: TStream);
    procedure SaveArrToStream(s: TStream); // = 1 if we're putting obstacles,
                  // = 2 if we're erasing them
  public
    StartPos : TPoint;     // starting position
    EndPos   : TPoint;     // the destination

    //Obstacle : TObstacleArray;
    Obstacle : Array of Array of Boolean;

    { draws obstacles }
    procedure ShowObstacles;

    { draws particular obstacle }
    procedure ShowObstacle(X,Y:Integer);

    { just puts square with specific color. wrong name actually :-) }
    procedure ShowStartEnd(X,Y,Color:Integer);

    { Path-Finding routine. returns True if successful }
    function FindPath:Boolean;
  end;

var
  MainForm: TMainForm;

implementation
Uses Pathes;
{$R *.DFM}

Var
 DirChangedPointsArr   :  Array of TDirChangedPoints;
 DirChangedPointsCount : integer;
 DirChangedPoints : TDirChangedPoints;

procedure TMainForm.FormCreate(Sender: TObject);
Var
i,j : Integer;
begin
 DimW:=MapW;
 DimH:=MapH;

 SetLength(Obstacle,MapH,MapW);

 For j :=0 to (DimH-1) do
  For i :=0 to(DimW-1) do
   Obstacle[j,i]:=false;

 StartPos:=Point(1,1);
 EndPos:=Point(MapW-1,MapH-1);
 ShowObstacles;
 Drawing:=0; // not drawing
end;


procedure TMainForm.SaveArrToStream(s: TStream);
Var
i,j : Integer;
begin
 S.Write(DimH,SizeOf(Integer));
 S.Write(DimW,SizeOf(Integer));

 For j :=0 to (DimH-1) do
  For i :=0 to(DimW-1) do
   S.Write(Obstacle[j,i],SizeOf(Boolean));
end;


procedure TMainForm.LoadArrFromStream(s: TStream);
Var
i,j : Integer;
begin
 s.Position:=0;
 S.Read(DimH,SizeOf(Integer));
 S.Read(DimW,SizeOf(Integer));
 SetLength(Obstacle,DimH,DimW);
 For j :=0 to (DimH-1) do
  For i :=0 to(DimW-1) do
   S.Read(Obstacle[j,i],SizeOf(Boolean));
end;

procedure TMainForm.ShowObstacle(X,Y:Integer);
begin
 Paint.Canvas.Pen.Color:=$FFFFFF;
 Paint.Canvas.Pen.Style:=psSolid;
 if (Obstacle[Y,X])=true then
  begin
   Paint.Canvas.Brush.Color:=$0000FF;
   Paint.Canvas.Brush.Style:=bsSolid;
  end else
  begin
   Paint.Canvas.Brush.Color:=$000000;
   Paint.Canvas.Brush.Style:=bsSolid;
  end;
 Paint.Canvas.Rectangle(X*dL,Y*dL,(X*dL)+dL+1,(Y*dL)+dL+1);
end;

procedure TMainForm.ShowStartEnd(X,Y,Color:Integer);
begin
 Paint.Canvas.Pen.Color:=$FFFFFF;
 Paint.Canvas.Pen.Style:=psSolid;
 Paint.Canvas.Brush.Color:=Color;
 Paint.Canvas.Brush.Style:=bsSolid;
 Paint.Canvas.Rectangle(X*dL,Y*dL,(X*dL)+dL+1,(Y*dL)+dL+1);
end;

procedure TMainForm.ShowObstacles;
Var X,Y:Integer;
begin
 for Y:=0 to MapH-1 do
  for X:=0 to MapW-1 do
   begin
    ShowObstacle(X,Y);
    if (X=StartPos.X)and(Y=StartPos.Y) then ShowStartEnd(X,Y,$FF7F00);
    if (X=EndPos.X)and(Y=EndPos.Y) then ShowStartEnd(X,Y,$FF00FF);
   end;
end;

procedure TMainForm.PaintPaint(Sender: TObject);
begin
 ShowObstacles;
end;

procedure TMainForm.PaintMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
Var Rx,Ry:Integer;
begin
 Rx := X div dL;
 Ry := Y div dL;

 if (Rx<0)or(Ry<0)or(Rx>MapW-1)or(Ry>MapH-1) then Exit;

 { putting/erasing obstacle  }
 if ObstacleButton.Down then
  begin
   { can't put obstacle on Start/End spot }
   if (Rx=StartPos.X)and(Ry=StartPos.Y) then Exit;
   if (Rx=EndPos.X)and(Ry=EndPos.Y) then Exit;

   if Button=mbLeft then Drawing:=1; // left mouse button - put obstacle }
   if Button=mbRight then Drawing:=2; // right mouse button - remove obstacle }

   if Drawing=1 then
    Obstacle[Ry,Rx]:=True
     else
    Obstacle[Ry,Rx]:=False;

   ShowObstacle(Rx,Ry); // showing current spot
  end;

 if StartButton.Down then
  begin
   if (Rx=EndPos.X)and(Ry=EndPos.Y) then Exit; // can't put Start position on End Position

   Obstacle[Ry,Rx]:=False; // removing obstacle from that spot
   ShowObstacle(StartPos.X,StartPos.Y); // hiding old Starting position

   StartPos.X:=Rx;
   StartPos.Y:=Ry;

   ShowStartEnd(Rx,Ry,$FF7F00); // drawing new Starting position
  end;
 if FinishButton.Down then
  begin
   if (Rx=StartPos.X)and(Ry=StartPos.Y) then Exit; // can't put Destination under Start position

   Obstacle[Ry,Rx]:=False; // removing obstacle from the spot
   ShowObstacle(EndPos.X,EndPos.Y); // hiding old End position

   EndPos.X:=Rx;
   EndPos.Y:=Ry;

   ShowStartEnd(Rx,Ry,$FF00FF); // drawign new End position
  end;
end;

procedure TMainForm.PaintMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 Drawing:=0; // user depressed button - stop drawing
end;

procedure TMainForm.PaintMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
Var Rx,Ry:Integer;
begin
 Rx:=X div dL;
 Ry:=Y div dL;

 if (Rx<0)or(Ry<0)or(Rx>MapW-1)or(Ry>MapH-1) then Exit;// Mouse is out of 20x20 bounds

 if Drawing=0 then Exit; // we aren't drawing anything

 if Drawing=1 then
  Obstacle[Ry,Rx]:=True
   else
  Obstacle[Ry,Rx]:=False;

 ShowObstacle(Rx,Ry);
end;

procedure TMainForm.PathButtonClick(Sender: TObject);
begin
 { executing path routine }
 if not FindPath then
  ShowMessage('Find path failed!');
end;


/////////////////////////////////////////////
function TMainForm.FindPath:Boolean;
var
 i,j,n  : Integer;
 PathInf      : TDXPath;

 procedure ShowDirections;
 Var
  i,j,xAdd,yAdd,Dir:integer;
 begin
 for j:=0 to MapH-1 do
  for i:=0 to MapW-1 do
   begin
    Dir:=PathInf.FDirMap[j,i]; // loading from direction map
    { drawing only if direction is set }
    if Dir<>255 then
     begin
      Dir:=NegDir[Dir]; // getting negative value

      { drawing offsets }
      xAdd:=DirX[Dir]*6;
      yAdd:=DirY[Dir]*6;

      Paint.Canvas.Pen.Style:=psSolid;
      Paint.Canvas.Pen.Color:=$FFFFFF;
      Paint.Canvas.MoveTo((i*dL)+ dL div 2,(j*dL)+ dL div 2);
      Paint.Canvas.LineTo((i*dL)+ dL div 2 + xAdd,(j*dL)+ dL div 2 + yAdd);
     end;
   end;
 end;

 procedure ShowPath;
 Var
  i,x,y:Integer;
  Dir : byte;
 begin
 Paint.Canvas.Brush.Color:=$3F9FFF;
 Paint.Canvas.Brush.Style:=bsSolid;

 for i:=0 to PathInf.FMaxPath-1 do
  begin
   x:=PathInf.FPath[i].x*dL;
   y:=PathInf.FPath[i].y*dL;
   Paint.Canvas.Rectangle(x,y,x+dL+1,y+dL+1);
   Dir:=PathInf.FDirMap[PathInf.FPath[i].y,PathInf.FPath[i].x];

   //Canvas.font.Color:=clRed;
   //Canvas.Brush.Style:=bsClear;
   //Canvas.TextOut(x,y,IntToStr(Dir));
  end;


 Paint.Canvas.Brush.Color:=clLime;

 for i:=0 to DirChangedPointsCount-1 do
 begin
   x:=DirChangedPointsArr[i].Point.x*dL;
   y:=DirChangedPointsArr[i].Point.y*dL;
   Paint.Canvas.Rectangle(x,y,x+dL+1,y+dL+1);
   Canvas.TextOut(x,y+10,IntToStr(i));
   Canvas.TextOut(x,y+20,IntToStr(DirChangedPointsArr[i].Dir));
 end;

 end;

//end local
begin
 PathInf:=TDXPath.Create(DimW,DimH);
  For j :=0 to (DimH-1) do
   For i :=0 to(DimW-1) do
    PathInf.FObstacle[j,i]:=Obstacle[j,i];

 PathInf.FStartPos:=StartPos;
 PathInf.FEndPos:=EndPos;

 mwFastTime1.Start;

 Result:=PathInf.FindPath;
 if Not Result then Exit;

 mwFastTime1.Stop;
 Label1.Caption:=mwFastTime1.ElapsedTime;

 DirChangedPointsCount:=PathInf.DirChangedPointsCount;
 SetLength(DirChangedPointsArr,DirChangedPointsCount);
 For n:=0 to DirChangedPointsCount-1 do
 begin
  DirChangedPointsArr[n]:=PathInf.DirChangedPointsArr[n];
 end;

 ShowObstacles;
 ShowDirections;
 ShowPath;
end;

procedure TMainForm.RefreshButtonClick(Sender: TObject);
begin
 ShowObstacles;
end;

procedure TMainForm.Button1Click(Sender: TObject);
Var
ms : TMemoryStream;
begin
ms:=TMemoryStream.Create;
try
 SaveArrToStream(ms);
 ms.SaveToFile(GetName('Map1.Dat'));
finally
ms.Free;
end;

end;

procedure TMainForm.Button2Click(Sender: TObject);
Var
ms : TMemoryStream;
begin
ms:=TMemoryStream.Create;
try
 ms.LoadFromFile(GetName('Map1.Dat'));
 LoadArrFromStream(ms);

finally
ms.Free;
end;

ShowObstacles;
end;

end.
